/*******************************************************************************

        @file ICU.d
        
        Copyright (c) 2004 Kris Bell
        
        This software is provided 'as-is', without any express or implied
        warranty. In no event will the authors be held liable for damages
        of any kind arising from the use of this software.
        
        Permission is hereby granted to anyone to use this software for any 
        purpose, including commercial applications, and to alter it and/or 
        redistribute it freely, subject to the following restrictions:
        
        1. The origin of this software must not be misrepresented; you must 
           not claim that you wrote the original software. If you use this 
           software in a product, an acknowledgment within documentation of 
           said product would be appreciated but is not required.

        2. Altered source versions must be plainly marked as such, and must 
           not be misrepresented as being the original software.

        3. This notice may not be removed or altered from any distribution
           of the source.

        4. Derivative works are permitted, but they must carry this notice
           in full and credit the original source.


                        ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


        @version        Initial version, October 2004      
        @author         Kris, John Reimer

*******************************************************************************/

module mango.icu.ICU;

private import std.string;

/*******************************************************************************

*******************************************************************************/

private static extern (C) uint strlen (char *s);
private static extern (C) uint wcslen (wchar *s);

/*******************************************************************************
        
        Some low-level routines to help bind the ICU C-API to D.

*******************************************************************************/

protected class ICU
{
        /***********************************************************************
        
                Use this for the primary argument-type to most ICU functions

        ***********************************************************************/

        protected typedef void* Handle;

        /***********************************************************************

                The library names to load within the target environment

        ***********************************************************************/

        version(Win32) 
        {
                protected static char[] icuuc = "icuuc30.dll";     
                protected static char[] icuin = "icuin30.dll";     
        }
        version(linux) 
        {
                protected static char[] icuuc = "libicuuc.so.30";
                protected static char[] icuin = "libicui18n.so.30";
        }

        /***********************************************************************

                The binary form of a version on ICU APIs is an array of 
                four bytes        

        ***********************************************************************/

        public struct   Version
                        {
                        ubyte[4] info;
                        }

        /***********************************************************************
        
                ICU error codes (the ones which are referenced)

        ***********************************************************************/

        protected enum  Error:int 
                        {
                        OK, 
                        BufferOverflow=15
                        }

        /***********************************************************************
        
        ***********************************************************************/

        protected static final bool isError (Error e)
        {
                return e > 0;
        }

        /***********************************************************************
        
        ***********************************************************************/

        protected static final void exception (char[] msg)
        {
                throw new ICUException (msg);
        }

        /***********************************************************************
        
        ***********************************************************************/

        protected static final void testError (Error e, char[] msg)
        {
                if (e > 0)
                    exception (msg);
        }

        /***********************************************************************
        
        ***********************************************************************/

        protected static final char* toString (char[] string)
        {
                static char[] empty = "";

                if (! string.length)
                      return ((cast(char*) string) ? empty : cast(char*) null);

                if (* (&string[0] + string.length))
                   {
                   // Need to make a copy
                   char[] copy = new char [string.length + 1];
                   copy [0..string.length] = string;
                   copy [string.length] = 0;
                   string = copy;
                   }
                return string;
        }

        /***********************************************************************
        
        ***********************************************************************/

        protected static final wchar* toString (wchar[] string)
        {
                static wchar[] empty = "";

                if (! string.length)
                      return ((cast(wchar*) string) ? empty : cast(wchar*) null);

                if (* (&string[0] + string.length))
                   {
                   // Need to make a copy
                   wchar[] copy = new wchar [string.length + 1];
                   copy [0..string.length] = string;
                   copy [string.length] = 0;
                   string = copy;
                   }
                return string;
        }

        /***********************************************************************
        
        ***********************************************************************/

        protected static final uint length (char* s)
        {
                return strlen (s);
        }

        /***********************************************************************
        
        ***********************************************************************/

        protected static final uint length (wchar* s)
        {
                return wcslen (s);
        }

        /***********************************************************************
        
        ***********************************************************************/

        protected static final char[] toArray (char* s)
        {
                if (s)
                    return s[0..strlen (s)];
                return null;
        }

        /***********************************************************************
        
        ***********************************************************************/

        protected static final wchar[] toArray (wchar* s)
        {
                if (s)
                    return s[0..wcslen (s)];
                return null;   
        }
}


/*******************************************************************************

*******************************************************************************/

class ICUException : Exception
{
        /***********************************************************************
        
                Construct exception with the provided text string

        ***********************************************************************/

        this (char[] msg)
        {
                super (msg);
        }
}


/*******************************************************************************
        
        Function address loader for Win32

*******************************************************************************/

version (Win32)
{
        /***********************************************************************

        ***********************************************************************/

        class FunctionLoader
        {
                private import std.c.windows.windows;

                /***************************************************************

                ***************************************************************/

                protected struct Bind
                {
                        void**  fnc;
                        char[]  name;      
                }

                /***************************************************************

                ***************************************************************/

                static final void* bind (char[] library, inout Bind[] targets)
                {
                        static char[] vrsn = "_3_0\0";

                        HANDLE lib = LoadLibraryA (ICU.toString(library));

                        foreach (Bind b; targets)
                                {
                                char[] name = b.name ~ vrsn;
                                *b.fnc = GetProcAddress (lib, name);
                                if (*b.fnc)
                                   {}// printf ("bound '%.*s'\n", name);
                                else
                                   throw new Exception ("required " ~ name ~ " in library " ~ library);
                                }
                        return lib;
                }

                /***************************************************************

                ***************************************************************/

                static final void unbind (void* library)
                {       
                        version (CorrectedTeardown)
                                 FreeLibrary (cast(HANDLE) library);
                }
        }
}


/*******************************************************************************
        
        2004-11-26:  Added Linux shared library support -- John Reimer

*******************************************************************************/

version (linux)
{
        // from include/bits/dlfcn.h on Linux
        const int RTLD_LAZY = 0x00001;     // Lazy function call binding
        const int RTLD_NOW = 0x00002;      // Immediate function call binding
        const int RTLD_NOLOAD = 0x00004;   // no object load
        const int RTLD_DEEPBIND = 0x00008;
        const int RTLD_GLOBAL = 0x00100;   // make object available to whole program    
        
        extern(C)
        {       
                void* dlopen(char* filename, int flag);
                char* dlerror();
                void* dlsym(void* handle, char* symbol);
                int   dlclose(void* handle);
        }
                
        class FunctionLoader
        {
                /***************************************************************

                ***************************************************************/

                protected struct Bind
                {
                        void**  fnc;
                        char[]  name;      
                }

                /***************************************************************

                ***************************************************************/

                static final void* bind (char[] library, inout Bind[] targets)
                {
                        static char[] vrsn = "_3_0\0";
                        static char[] errorInfo;
                        // printf("the library is %s\n", ICU.toString(library));
                        void* lib = dlopen(ICU.toString(library), RTLD_NOW);
                        
                        // clear the error buffer
                        dlerror();

                        foreach (Bind b; targets)
                        {
                                char[] name = b.name ~ vrsn;
                                
                                *b.fnc = dlsym (lib, name);
                                if (*b.fnc)
                                   {}// printf ("bound '%.*s'\n", name);
                                else {
                                        // errorInfo = ICU.toArray(dlerror());
                                        // printf("%s", dlerror());
                                        throw new Exception ("required " ~ name ~ " in library " ~ library);
                                }
                        }
                        return lib;
                }

                /***************************************************************

                ***************************************************************/

                static final void unbind (void* library)
                {       
                        version (CorrectedTeardown)
                                {
                                if (! dlclose (library))
                                      throw new Exception ("close library failed\n");       
                                }
                }
        }
}
