/*******************************************************************************

        @file FileSystem.d
        
        Copyright (c) 2004 Kris Bell
        
        This software is provided 'as-is', without any express or implied
        warranty. In no event will the authors be held liable for damages
        of any kind arising from the use of this software.
        
        Permission is hereby granted to anyone to use this software for any 
        purpose, including commercial applications, and to alter it and/or 
        redistribute it freely, subject to the following restrictions:
        
        1. The origin of this software must not be misrepresented; you must 
           not claim that you wrote the original software. If you use this 
           software in a product, an acknowledgment within documentation of 
           said product would be appreciated but is not required.

        2. Altered source versions must be plainly marked as such, and must 
           not be misrepresented as being the original software.

        3. This notice may not be removed or altered from any distribution
           of the source.

        4. Derivative works are permitted, but they must carry this notice
           in full and credit the original source.


                        ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


        @version        Initial version, March 2004      
        @author         Kris, John Reimer


*******************************************************************************/

module mango.io.FileSystem;

private import  mango.io.Utf8,
                mango.io.FilePath,
                mango.io.Exception;

// add some externs for linux
version (Posix)
{
        extern (C) int strlen (char *);
}

/*******************************************************************************

        Models an OS-specific file-system. Included here are methods to 
        list the system roots ("C:", "D:", etc) and to manipulate the
        current working directory.

*******************************************************************************/

class FileSystem
{
        /***********************************************************************
        
                A set of file-system specific constants for file and path
                seperators (chars and strings).

        ***********************************************************************/

        version (Win32)
        {
                static const char PathSeperatorChar = '\\';
                static const char FileSeperatorChar = '.';
                static const char RootSeperatorChar = ':';

                static const char[] PathSeperatorString = "\\";
                static const char[] FileSeperatorString = ".";
                static const char[] RootSeperatorString = ":";

                static const char[] NewlineString = "\r\n";
        }
        
        
        version (Posix)
        {
                static const char PathSeperatorChar = '/';
                static const char FileSeperatorChar = '.';
                static const char RootSeperatorChar = ':';

                static const char[] PathSeperatorString = "/";
                static const char[] FileSeperatorString = ".";
                static const char[] RootSeperatorString = ":";

                static const char[] NewlineString = "\n";
        }            

        version (Win32)
        {
                private import std.c.windows.windows;

                /***********************************************************************
                        
                        List the set of root devices (C:, D: etc)

                        @todo not currently implemented.

                ***********************************************************************/

                static char[][] listRoots ()
                {
                        assert(0);
                        return null;
                }

                /***********************************************************************

                        Set the current working directory

                ***********************************************************************/

                static void setDirectory (FilePath fp)
                {
                        if (! SetCurrentDirectoryW (fp.toUtf16))
                              throw new IOException ("Failed to set current directory");
                }

                /***********************************************************************

                        Get the current working directory
                
                ***********************************************************************/

                static FilePath getDirectory ()
                {
                        wchar[] dir;
                        int     length;

                        length = GetCurrentDirectoryW (0, null);
                        if (length)
                           {
                           dir = new wchar [length];
                           GetCurrentDirectoryW (length, dir);
                           return new FilePath (Utf8.encode (dir));
                           }
                        throw new IOException ("Failed to get current directory");
                }
        }
        
        
        version (Posix)
        {
                private import std.c.linux.linux;

                /***********************************************************************

                        List the set of root devices.

                        @todo not currently implemented.

                ***********************************************************************/

                static char[][] listRoots ()
                {
                        assert(0);
                        return null;
                }

                /***********************************************************************

                        Set the current working directory

                ***********************************************************************/

                static void setDirectory (FilePath fp)
                {
                        if (std.c.linux.linux.chdir (fp.toUtf8))
                            throw new IOException ("Failed to set current directory");
                }

                /***********************************************************************

                        Get the current working directory
                
                ***********************************************************************/

                static FilePath getDirectory ()
                {
                        char *s = std.c.linux.linux.getcwd (null, 0);
                        if (s) 
                            // dup the string so we can hang onto it                            
                            return new FilePath (s[0..strlen(s)].dup);

                        throw new IOException ("Failed to get current directory");
                }
        }            
}
