/*******************************************************************************

        @file Token.d
        
        Copyright (c) 2004 Kris Bell
        
        This software is provided 'as-is', without any express or implied
        warranty. In no event will the authors be held liable for damages
        of any kind arising from the use of this software.
        
        Permission is hereby granted to anyone to use this software for any 
        purpose, including commercial applications, and to alter it and/or 
        redistribute it freely, subject to the following restrictions:
        
        1. The origin of this software must not be misrepresented; you must 
           not claim that you wrote the original software. If you use this 
           software in a product, an acknowledgment within documentation of 
           said product would be appreciated but is not required.

        2. Altered source versions must be plainly marked as such, and must 
           not be misrepresented as being the original software.

        3. This notice may not be removed or altered from any distribution
           of the source.

        4. Derivative works are permitted, but they must carry this notice
           in full and credit the original source.


                        ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


        @version        Initial version, March 2004      
                        Circular dependency split; Oct 2004  

        @author         Kris, Chris Sauls


*******************************************************************************/

module mango.io.Token;

private import  std.ctype;

private import  mango.io.Exception;

private import  mango.io.model.IWriter;

/*******************************************************************************

        This provides backward compatability by importing the additional
        token styles into this namespace.

*******************************************************************************/

public  import  mango.io.TokenEx;


/*******************************************************************************

        Tokens used by Tokenizer class. Tokens do not copy their content
        so they are quite useful for parsing quantites of data quickly. 
        Conversely since each token is mapped into an external buffer, 
        you should be aware that changes to said buffer will impact any
        tokens based upon it. You may sidestep this by using the clone()
        method, or toString().dup

        Tokens can convert from a variety of numeric format to ascii text.
        Formats currently include int, uint, long, ulong, and real. Each
        number may be preceded by whitespace, and an optional '+' or '-'
        specifier. Note that real-number format is simplistic in that it 
        does not support exponential declarations.  Note the conversion 
        methods should probably be moved elsewhere.

        Here's a brief example of how to apply Token with Tokenizers:

        @code
        // open a file for reading
        FileConduit fc = new FileConduit ("test.txt");

        // create a buffer for reading the file
        IBuffer buffer = fc.createBuffer ();

        // create a token for receiving the line
        Token token = new Token;

        // read file a line at a time. Method next() returns false when no more 
        // delimiters are found. Note there may be an unterminated line at eof
        while (Tokenizers.line.next(buffer, token) || token.getLength())
               Stdout.put(token).cr();
        @endcode

        See also BoundToken, ReaderToken, CompositeToken and HybridToken.

*******************************************************************************/

class Token : IWritable
{ 
        private int     type;
        private char[]  content;

        private static TokenException Error;

        private enum State {Begin, Number, Fraction};

        alias void delegate(int digit) Accumulator;

        /***********************************************************************
        
        ***********************************************************************/

        private static  real divisors[] = 
                        [
                        0.1, 
                        0.01, 
                        0.001, 
                        0.0001, 
                        0.00001, 
                        0.000001, 
                        0.0000001, 
                        0.00000001, 
                        0.000000001, 
                        0.0000000001, 
                        0.00000000001, 
                        0.000000000001, 
                        0.0000000000001, 
                        0.00000000000001, 
                        0.000000000000001, 
                        0.0000000000000001, 
                        0.00000000000000001,
                        0.000000000000000001,
                        0.0000000000000000001,
                        0.00000000000000000001,
                        0.000000000000000000001,
                        0.0000000000000000000001,
                        ];

        /***********************************************************************
        
                Construct a static exception instance.

        ***********************************************************************/

        static this()
        {
                Error = new TokenException ("unexpected numeric format");
        }

        /***********************************************************************
        
                Set the content of this token.

        ***********************************************************************/

        Token set (char[] content)
        {
                this.content = content;
                return this;
        }

        /***********************************************************************
                
                Return the length of this token.

        ***********************************************************************/

        int getLength ()
        {
                return content.length;
        }

        /***********************************************************************
        
                Set the type of this token. Token types can be useful when
                one wishes to categorize input patterns.

        ***********************************************************************/

        Token setType (int type)
        {
                this.type = type;
                return this;
        }

        /***********************************************************************
        
                Return the type associated with this token. See setType().

        ***********************************************************************/

        int getType ()
        {
                return type;
        }

        /***********************************************************************
        
                Convert this token to an integer.

        ***********************************************************************/

        int toInt ()
        {
                int i = 0;

                void multiply (int digit)
                {
                        i = i * 10 + digit;
                }

                return parseDigits (&multiply, null) ? -i : i;
        }

        /***********************************************************************
        
                Convert this token to an unsigned integer.

        ***********************************************************************/

        uint toUInt ()
        {
                uint i = 0;

                void multiply (int digit)
                {
                        i = i * 10 + digit;
                }

                if (parseDigits (&multiply, null))
                    throw Error;
                return i;
        }

        /***********************************************************************
        
                Convert this token to a long integer.

        ***********************************************************************/

        long toLong ()
        {
                long l = 0;

                void multiply (int digit)
                {
                        l = l * 10 + digit;
                }

                return parseDigits (&multiply, null) ? -l : l;
        }

        /***********************************************************************
        
                Convert this token to an unsigned long integer.

        ***********************************************************************/

        ulong toULong ()
        {
                ulong l = 0;

                void multiply (int digit)
                {
                        l = l * 10 + digit;
                }

                if (parseDigits (&multiply, null))
                    throw Error;
                return l;
        }

        /***********************************************************************

                Convert this token to a real.

        ***********************************************************************/

        real toReal ()
        {
                real    r = 0,
                        f = 0;
                int     divisor = 0;

                void whole (int c)
                {
                        r = r * 10.0 + c;
                }

                void fraction (int c)
                {                        
                        f = f * 10.0 + c;
                        ++divisor;
                }

                bool neg = parseDigits (&whole, &fraction);

                if (divisor)
                    if (divisor < divisors.length)
                        r += f * divisors [divisor-1];
                    else
                       throw Error;

                return neg ? -r : r;
        }

        /***********************************************************************
        
                Clone this token, making a copy of the content also.

        ***********************************************************************/

        Token clone ()
        {
                Token clone = new Token;

                clone.set (toString (false));
                clone.type = type;
                return clone;
        }

        /***********************************************************************
        
                Return a reference to this tokens content. Duplicate it
                only if 'slice' is explicitly set to false (defaults to 
                a slice instead).

        ***********************************************************************/

        char[] toString (bool slice = true)
        {
                if (slice)
                    return content;
                return content.dup;
        }

        /***********************************************************************
        
                Is this token equal to another?

        ***********************************************************************/

        override int opEquals (Object o)
        {
                Token other = cast(Token) o;

                if (other is null)
                    return super.opEquals (o);
                return typeid(char[]).equals (&content, &other.content);
                //content.typeinfo.equals (content, other.content);
        }

        /***********************************************************************
        
                Compare this token to another.

        ***********************************************************************/

        override int opCmp (Object o)
        {
                Token other = cast(Token) o;

                if (other is null)
                    return super.opCmp (o);

                return typeid(char[]).compare (&content, &other.content);
                //content.typeinfo.compare (content, other.content);
        }

        /***********************************************************************
        
                Hash this token

        ***********************************************************************/

        override uint toHash ()
        {
                return typeid(char[]).getHash (&content);
                // content.typeinfo.getHash (content);
        }

        /***********************************************************************
        
                Make the Token class compatible with IWriter instances.

        ***********************************************************************/

        void write (IWriter w)
        {
                w.put (content);
        }

        /***********************************************************************
        
                Parse a set of digits, optionally prefixed by a '+' or '-', 
                and optionally with a '.' contained within. Delegates come
                in quite handy here, and add relatively little overhead. 

        ***********************************************************************/

        private bool parseDigits (Accumulator dg, Accumulator fraction)
        {
                bool            neg = false;
                State           state = State.Begin; 

                foreach (char c; content)
                         switch (state)   
                                {
                                case State.Begin:
                                     if (isspace (c))
                                         break;

                                     state = State.Number;
                                     if (c == '-')
                                        {
                                        neg = true;
                                        break;
                                        }
                                     else
                                        if (c == '+')
                                            break;
                                     // fall thru!
  
                                case State.Number:
                                     if (c == '.' && fraction)
                                        {
                                        state = State.Fraction;
                                        dg = fraction;
                                        break;
                                        }
                                     // fall thru!

                                case State.Fraction:
                                     if (isdigit (c))
                                         dg (c - '0');
                                     else
                                        throw Error;
                                     break;

                                default:
                                }
                return neg;
        }
}
