/*******************************************************************************

        @file IWriter.d
        
        Copyright (c) 2004 Kris Bell
        
        This software is provided 'as-is', without any express or implied
        warranty. In no event will the authors be held liable for damages
        of any kind arising from the use of this software.
        
        Permission is hereby granted to anyone to use this software for any 
        purpose, including commercial applications, and to alter it and/or 
        redistribute it freely, subject to the following restrictions:
        
        1. The origin of this software must not be misrepresented; you must 
           not claim that you wrote the original software. If you use this 
           software in a product, an acknowledgment within documentation of 
           said product would be appreciated but is not required.

        2. Altered source versions must be plainly marked as such, and must 
           not be misrepresented as being the original software.

        3. This notice may not be removed or altered from any distribution
           of the source.

        4. Derivative works are permitted, but they must carry this notice
           in full and credit the original source.


                        ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


        @version        Initial version, March 2004      
        @author         Kris, Ivan Senji (the "alias opShl put" idea)


*******************************************************************************/

module mango.io.model.IWriter;

public import mango.io.model.IBuffer;

/*******************************************************************************

*******************************************************************************/

typedef void delegate(void* dst, uint count) BufferEncoder;

/*******************************************************************************

        Interface to make any class compatible with any IWriter.

*******************************************************************************/

interface IWritable
{
        void write (IWriter w);
}

/*******************************************************************************

        Any class implementing IEncoder can be bound to a writer using
        the bind() method.
        
*******************************************************************************/

interface IEncoder
{
        ConverterType type ();

        BufferEncoder bind (IWriter writer);
}

/*******************************************************************************

        Make a signature for IWritable classes to use when they wish to 
        avoid processing by writers such as TokenWriter.

*******************************************************************************/

interface IPhantomWriter : IWritable
{
}

/*******************************************************************************

        Make a signature for Newline classes

*******************************************************************************/

interface INewlineWriter : IPhantomWriter
{
}

/*******************************************************************************

        IWriter interface. IWriter provides the means to append formatted 
        data to an IBuffer, and exposes a convenient method of handling a
        variety of data types. In addition to writing native types such
        as integer and char[], writers also process any class which has
        implemented the IWritable interface (one method).

*******************************************************************************/

interface IWriter
{
        alias opShl put;

        /***********************************************************************
        
                These are the basic writer methods

        ***********************************************************************/

        IWriter opShl (bool x);
        IWriter opShl (ubyte x);
        IWriter opShl (byte x);
        IWriter opShl (ushort x);
        IWriter opShl (short x);
        IWriter opShl (uint x);
        IWriter opShl (int x);
        IWriter opShl (ulong x);
        IWriter opShl (long x);
        IWriter opShl (float x);
        IWriter opShl (double x);
        IWriter opShl (real x);
        IWriter opShl (char x);
        IWriter opShl (wchar x);
        IWriter opShl (dchar x);

        IWriter opShl (byte[] x);
        IWriter opShl (short[] x);
        IWriter opShl (int[] x);
        IWriter opShl (long[] x);
        IWriter opShl (ubyte[] x);
        IWriter opShl (ushort[] x);
        IWriter opShl (uint[] x);
        IWriter opShl (ulong[] x);
        IWriter opShl (float[] x);
        IWriter opShl (double[] x);
        IWriter opShl (real[] x);
        IWriter opShl (char[] x);

        // these two will have to wait until Walter fixes 
        // the string-literal conversion fiasco ...
        IWriter opShlw (wchar[] x);
        IWriter opShld (dchar[] x);

        IWriter opShl (IWritable x);

        /***********************************************************************
        
                Bind an IEncoder to the writer. Encoders are intended to
                be used as a conversion mechanism between various character
                representations (encodings), or the translation of any data
                type from one representation to another. Each data type may
                be configured with a distinct encoder, covering all native
                types (15 in total).

                An appropriate encoder set should be attached to each 
                IWriter, and thus be available for subsequent use. A raw 
                binary implementation is attached by default (no encoding).

                See module mango.icu.UMango for an example of encoder 
                implementation -- those classes bind the ICU converters 
                to this IO package.

        ***********************************************************************/

        void setEncoder (IEncoder e); 

        /***********************************************************************

                Output a newline. Do this indirectly so that it can be 
                intercepted by subclasses.
        
        ***********************************************************************/

        IWriter cr ();

        /***********************************************************************
        
                Flush the output of this writer. Throws an IOException 
                if the operation fails.

        ***********************************************************************/

        IWriter flush ();

        /***********************************************************************
        
                Return the associated buffer

        ***********************************************************************/

        IBuffer getBuffer ();
}

