/*******************************************************************************

        @file Text.d
        
        Copyright (c) 2004 Kris Bell
        
        This software is provided 'as-is', without any express or implied
        warranty. In no event will the authors be held liable for damages
        of any kind arising from the use of this software.
        
        Permission is hereby granted to anyone to use this software for any 
        purpose, including commercial applications, and to alter it and/or 
        redistribute it freely, subject to the following restrictions:
        
        1. The origin of this software must not be misrepresented; you must 
           not claim that you wrote the original software. If you use this 
           software in a product, an acknowledgment within documentation of 
           said product would be appreciated but is not required.

        2. Altered source versions must be plainly marked as such, and must 
           not be misrepresented as being the original software.

        3. This notice may not be removed or altered from any distribution
           of the source.

        4. Derivative works are permitted, but they must carry this notice
           in full and credit the original source.


                        ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

        
        @version        Initial version, April 2004      
        @author         Kris


*******************************************************************************/

module mango.utils.Text;

private import  std.date,
                std.ctype;

private import  std.c.stdio;

/******************************************************************************

******************************************************************************/

extern (C)
{
        char* memchr (char *, char, uint);
        int   memcmp (char *, char *, uint);
}       

/******************************************************************************

        Placeholder for a variety of wee functions. Some of these are
        handy for Java programmers, but the primary reason for their
        existance is that they don't allocate memory (unlike Phobos).
        That is, processing is performed in-place or within a client-
        provided temporary buffer.

******************************************************************************/

class Text
{
        /**********************************************************************

                Replace all instances of one char with another (in place)

        **********************************************************************/

        final static char[] replace (char[] source, char match, char replacement)
        {
                char *p;
                char *scan = &source[0];
                int length = source.length;

                while ((p = memchr (scan, match, length)) != null)
                      {
                      *p = replacement;
                      length -= (p - scan);
                      scan = p;
                      }
                return source;
        }

        /**********************************************************************

                Return the index of the first instance of 'match'

        **********************************************************************/

        final static int indexOf (char[] source, char match)
        {
                return indexOf (source, match, 0);
        }

        /**********************************************************************

                Return the index of the first instance of 'match', starting
                at position 'start'
                
        **********************************************************************/

        final static int indexOf (char[] source, char match, int start)
        {
                if (start < source.length)
                   {
                   char *p = memchr (&source[start], match, source.length - start);
                   if (p)
                       return start + (p - &source[start]);
                   }
                return -1;
        }

        /**********************************************************************

                Return the index of the first instance of 'match'

        **********************************************************************/

        final static int indexOf (char[] source, char[] match)
        {
                return indexOf (source, match, 0);
        }

        /**********************************************************************

                Return the index of the first instance of 'match', starting
                at position 'start'
                
        **********************************************************************/

        final static int indexOf (char[] source, char[] match, int start)
        {
                int length = match.length;
                int extent = source.length - length + 1;
                
                for (; start < extent; ++start)
                    {
                    start = indexOf (source, match[0], start);
                    if (start < 0)
                        break;
                    else
                       if (memcmp (&source[start], match, length) == 0)
                           return start;
                    }
                return -1;
        }

        /**********************************************************************

                Trim the provided string by stripping whitespace from 
                both ends. Returns a slice of the original content.

        **********************************************************************/

        final static char[] trim (char[] source)
        {
                if (source.length)
                   {
                   int  front,
                        back = source.length;

                   while (front < back && isspace(source[front]))
                          ++front;

                   while (back > front && isspace(source[back-1]))
                          --back;
                   
                   if (front > 0 || back < source.length)   
                       return source[front..back];
                   } 
                return source;
        }

        /**********************************************************************

                Convert a string into a base-10 number.

        **********************************************************************/


        final static int atoi (char[] digits)
        {
                return atoi (digits, 10);
        }

        /**********************************************************************

                Convert a string into a base-X number where X is specified
                by 'radix'

        **********************************************************************/

        final static int atoi (char[] digits, int radix)
        {
                int i;

                foreach (char c; digits)
                        {
                        if (isspace(c))
                            continue;

                        if (c >= 'a' && c <= 'f')
                            c -= 39;
                        else
                           if (c >= 'A' && c <= 'F')
                               c -= 7;
                           else
                              if (c < '0' || c > '9')
                                  throw new Exception ("invalid number '"~digits~"'");

                        i = i * radix + c - '0';
                        }
                return i;
        }


        /**********************************************************************

                Convert an integer to a string, without allocating space.

        **********************************************************************/

        final static char[] itoa (char[] s, long i)
        in {
           assert (s.length > 0);
           }
        body 
        {
                int len = s.length;
                do {
                   char c = i % 10 + '0';
                   s[--len] = c;
                   i /= 10;
                   } while (i && len);
                return s[len..s.length];
        }

        /**********************************************************************

                Convert an integer to a string

        **********************************************************************/

        final static char[] toString (int i)
        {
                char t[10];

                return itoa (t, i).dup;
        }

        /**********************************************************************

                Convert a string to a date. This is just a symmetrical 
                wrapper around the Phobos functionality.
               
        **********************************************************************/

        final static long atod (char[] date)
        {
                return parse (date);
        }

        /**********************************************************************

                Shamelessly plundered from std.date (with permission), this 
                version avoids allocating memory.

                Returns a populated slice of the provided buffer; with zero
                length if the date was invalid.

        **********************************************************************/

        final static char[] dtoa (char[] buffer, long time)
        {
                int     hr;
                int     mn;
                int     len;
                char    sign;
                d_time  t,
                        dst,
                        offset;

                // Years are supposed to be -285616 .. 285616, or 7 digits
                // "Tue Apr 02 02:04:57 GMT-0800 1996"
                assert(buffer.length >= 29 + 7 + 1);

                if (time == d_time_nan)
                    return buffer[0..0];

                dst = DaylightSavingTA (time);
                offset = LocalTZA + dst;
                t = time + offset;
                sign = '+';
                if (offset < 0)
                   {
                   sign = '-';
                   offset = -(LocalTZA + dst);
                   }

                mn = cast(int)(offset / msPerMinute);
                hr = mn / 60;
                mn %= 60;

                len = sprintf (buffer, "%.3s %.3s %02d %02d:%02d:%02d GMT%c%02d%02d %d",
                               &daystr[WeekDay(t) * 3],
                               &monstr[MonthFromTime(t) * 3],
                               DateFromTime(t),
                               cast(int)HourFromTime(t), cast(int)MinFromTime(t), 
                               cast(int)SecFromTime(t), sign, hr, mn,
                               cast(long)YearFromTime(t));

                return buffer[0..len];
        }

        /**********************************************************************

                in-place conversion to lowercase 

        **********************************************************************/

        final static char[] tolower (inout char[] src)
        {
                foreach (int i, char c; src)
                         if (c >= 'A' && c <= 'Z')
                             src[i] = c + ('a' - 'A');
                return src;
        }

}

